#!/usr/bin/perl
# train the svm

use Getopt::Long;

#
# read command line and check its correctness
#
$model = "";
$tests = "";
$testf = "";
$seqlen = 0;
$feats = "";
$out = "";
$fsel = "y";
$thresh = 0.0;
$progdir = ".";
$scrdir = "";
$flanklen = 3;
$ptsize = 20000;
$pval = "y";
$npval = 10000;
$fpval = "";
$spval = "";

$options = "Usage: ./PREDICT.pl <options>
  REQUIRED INPUTS:
   -m  model prefix (required)
   -tf test sequences in fasta format (test sequence length will be inferred)
   -l  length of test sequences (positive integer, required for -tf option)
   -ts test sequences, listed as literal sequences
       NOTE: one of -tf or -ts is required
   -f  feature list (required)
   -o  output prefix (required)

  EXTRA SETTINGS:
   -s  feature selection (y or n, default $fsel)
       note that the use of this option must be consistent with how the training was performed
   -r  threshold for calling positive hits (real number, default $thresh)

  STATISTICAL SIGNFICANCE CALCULATIONS:
   -pv compute p-values for each binding site (y or n, default $pval)
       two methods are available (default is (1)):
	 (1) Monte Carlo: generation of random sequences with the same GC-content
	     as your test sequences (use -pn option to change number), or
	 (2) input background sequences with -pf or -ps
   -pn number of background sequences to use (default $npval)
   -pf background sequences in fasta format (-l option must be used)
   -ps background sequences, listed as literal sequences

  COMPUTATIONAL SETTINGS (won't affect results, but possibly performance):
   -b  split size: size of test sequence subset to do at a time (positive integer, default $ptsize)
       making this too big may overload your memory, making it too small may increase overall time
   -p  program directory (for batch running on a cluster)
       set to be the directory that contains all the scripts (default: ./)
   -t  temporary scratch directory (for batch running on a cluster)
       set as the location to write out partial predictions (may need several GB free)
       ideally should be local to the compute node, as intermediate files are large (default: same directory as output)
";


GetOptions('m=s' => \$model,
           'tf=s' => \$testf,
           'ts=s' => \$tests,
           'l=i' => \$seqlen,
           'f=s' => \$feats,
           'o=s' => \$out,
           's=s' => \$fsel,
           'r=f' => \$thresh,
           'p=s' => \$progdir,
	   't=s' => \$scrdir,
	   'pv=s' => \$pval,
	   'pn=i' => \$npval,
	   'pf=s' => \$fpval,
	   'ps=s' => \$spval,
           'b=i' => \$ptsize) || die "$options\n";

if( $model eq "" ){
	print "Error: set model prefix\n$options\n";
	exit;
}
if( $tests eq "" && $testf eq "" ){
	print "Error: set test sequences\n$options\n";
	exit;
}
if( $testf ne "" && $seqlen <= 0 ){
	print "Error: set sequence length if using fasta format\n$options\n";
	exit;
}
if( $testf ne "" && $tests ne "" ){
	print "Error: using only one of the -tf and -ts options\n$options\n";
	exit;
}
if( $feats eq "" ){
	print "Error: set feature list\n$options\n";
	exit;
}
if( $out eq "" ){
	print "Error: set output prefix\n$options\n";
	exit;
}
if( $npval < 1000 ){
	print "Error: set -pn to be fairly large (at least 1000)\n$options\n";
	exit;
}
if( $spval ne "" && $fpval ne "" ){
	print "Error: set only one of -pf or -ps\n$options\n";
	exit;
}
if( $fpval ne "" && $seqlen <= 0 ){
	print "Error: set sequence length if using fasta format for background sequences\n$options\n";
	exit;
}

printf("\nInput model: %s\nTest sequences: %s%s\nFeatures: %s\nOutput: %s\n", $model, $tests, $testf, $feats, $out);
if( $fsel ne "y" ){
	if( $fsel ne "n" ){
		print "Error: set feature selection to be y or n\n";
		exit;
	}
	printf("Not using feature selection\n");
}
if( $pval eq "y" ){
	if( $fpval eq "" && $spval eq "" ){
		print "Computing p-values by generating $npval random sequences\n";
	}
	else{
		print "Computing p-values using $fpval$spval as background\n";
	}
}

#
# make directory to store temporary outputs
#
@parts = split('/',$out);
$pref = $parts[$#parts];
if( $scrdir eq "" ){
	if( $#parts > 0 ){
		$scrdir = "$parts[0]";
		if( substr($out,0,1) eq "/" ){
			$scrdir = "";
		}
		for($i=1; $i< $#parts; $i++){
			$scrdir = $scrdir."/".$parts[$i];
		}
	}
	else{
		$scrdir = ".";
	}
}
$check = `[ -d $scrdir/$pref\_parts ] && echo "y" || echo "n"`;
chomp($check);
if( $check eq "n" ){
	system("mkdir $scrdir/$pref\_parts");
}


#
# set value of $test according to format of test sequences
#
if( $tests ne "" ){
	$test = $tests;
	$bedcheck = "n";
}
else{
	system("$progdir/fasta_to_seqs.pl $testf $out $seqlen > $out.log");
	&check_stat("fasta_to_seqs.pl");
	$test = "$out.seq";
	$bedcheck = `[ -f $out.bed ] && echo "y" || echo "n"`;
	chomp($bedcheck);
}

#
# set value of $testp according to p-value inputs
#
if( $pval eq "y" ){
	if( $spval eq "" && $fpval eq "" ){
		$testp = "$scrdir/$pref.bk.seq";
		$gc = `$progdir/count_gc.pl $test`;
		chomp($gc);
		
		# get seqlen from test
		open(IN,"$test");
		$line = <IN>;
		chomp($line);
		$seqlen = length($line);
		close(IN);

		system("$progdir/make_random_seqs.pl $npval $gc $seqlen $testp");
	}
	elsif( $spval ne "" ){
		$testp = $spval;
	}
	else{
		system("$progdir/fasta_to_seqs.pl $fpval $out.bkgrnd $seqlen > $out.log");
		&check_stat("fasta_to_seqs.pl");
		$testp = "$out.bkgrnd.seq";
	}

	#
	# in increments of $ptsize sequences, parameterize and predict the test squences
	#
	print "\nMaking background predictions\n";
	$len = `wc -l $testp | awk '{print \$1}'`;
	chomp($len);
	$ftomake = sprintf("%i",$len/$ptsize);
	if( $len%$ptsize != 0 ){
		$ftomake++;
	}
	printf("   Splitting %i background sequences in %i parts of size %i\n", $len, $ftomake, $ptsize);
	open(TSTS,"$testp");
	$i=1;
	$j=1;
	open(PRT,">$scrdir/$pref\_parts/$pref-tmp.bk.seq");
	while( $line = <TSTS> ){
		print PRT "$line";
		if( $j%$ptsize == 0 || $j == $len ){
			print "\r   Parameterizing and predicting $i of $ftomake   ";
			close(PRT); 
			system("$progdir/parameterize.pl $scrdir/$pref\_parts/$pref-tmp.bk.seq $scrdir/$pref\_parts/$pref-tmp.bk.vec $flanklen $feats $model.trainvec.maxes $out.log");
			&check_stat("parameterize.pl");
			if( $fsel eq "y" ){system("$progdir/shorten_features.pl $scrdir/$pref\_parts/$pref-tmp.bk.vec $model.vecvals $scrdir/$pref\_parts/$pref-tmp.bk.shortvec");}
			else{system("mv $scrdir/$pref\_parts/$pref-tmp.bk.vec $scrdir/$pref\_parts/$pref-tmp.bk.shortvec");}
			system("$progdir/svm-predict $scrdir/$pref\_parts/$pref-tmp.bk.shortvec $model.model $scrdir/$pref\_parts/$pref-$i.bk.predict > $scrdir/$pref.log");
	                $flen = `wc -l $scrdir/$pref\_parts/$pref-$i.bk.predict | awk '{print \$1}'`;
	                chomp($flen);
	                system("head -$flen $scrdir/$pref.log > $scrdir/$pref\_parts/$pref-$i.bk.scores");
			$i++;
			if( $i<= $ftomake ){
				open(PRT,">$scrdir/$pref\_parts/$pref-tmp.bk.seq");
			}
		}
		$j++;
	}
	
	#
	# concatenate the .predict results
	#
	print "\n\nConcatenating and sorting the background scores\n";
	$allsc = "";
	for($i=1; $i<= $ftomake; $i++){
	        $allsc = $allsc." $scrdir/$pref\_parts/$pref-$i.bk.scores";
	}
	system("cat $allsc > $out.bk.scores");
}

#
# in increments of $ptsize sequences, parameterize and predict the test squences
#
print "\nMaking predictions\n";
$len = `wc -l $test | awk '{print \$1}'`;
chomp($len);
$ftomake = sprintf("%i",$len/$ptsize);
if( $len%$ptsize != 0 ){
	$ftomake++;
}
printf("   Splitting %i test sequences in %i parts of size %i\n", $len, $ftomake, $ptsize);
open(TSTS,"$test");
$i=1;
$j=1;
open(PRT,">$scrdir/$pref\_parts/$pref-tmp.seq");
while( $line = <TSTS> ){
	print PRT "$line";
	if( $j%$ptsize == 0 || $j == $len ){
		print "\r   Parameterizing and predicting $i of $ftomake   ";
		close(PRT); 
		system("$progdir/parameterize.pl $scrdir/$pref\_parts/$pref-tmp.seq $scrdir/$pref\_parts/$pref-tmp.vec $flanklen $feats $model.trainvec.maxes $out.log");
		&check_stat("parameterize.pl");
		if( $fsel eq "y" ){system("$progdir/shorten_features.pl $scrdir/$pref\_parts/$pref-tmp.vec $model.vecvals $scrdir/$pref\_parts/$pref-tmp.shortvec");}
		else{system("mv $scrdir/$pref\_parts/$pref-tmp.vec $scrdir/$pref\_parts/$pref-tmp.shortvec");}
		system("$progdir/svm-predict $scrdir/$pref\_parts/$pref-tmp.shortvec $model.model $scrdir/$pref\_parts/$pref-$i.predict > $scrdir/$pref.log");
                $flen = `wc -l $scrdir/$pref\_parts/$pref-$i.predict | awk '{print \$1}'`;
                chomp($flen);
                system("head -$flen $scrdir/$pref.log > $scrdir/$pref\_parts/$pref-$i.scores");
		$i++;
		if( $i<= $ftomake ){
			open(PRT,">$scrdir/$pref\_parts/$pref-tmp.seq");
		}
	}
	$j++;
}

#
# concatenate the .predict results; assign p-values if appropriate
#
print "\n\nConcatenating the split-up test predictions and cleaning up\n";
$all = "";
$allsc = "";
for($i=1; $i<= $ftomake; $i++){
        $all = $all." $scrdir/$pref\_parts/$pref-$i.predict";
        $allsc = $allsc." $scrdir/$pref\_parts/$pref-$i.scores";
}
system("cat $all > $scrdir/$pref.predict");
system("cat $allsc > $scrdir/$pref.scores");

if( $bedcheck eq "n" ){
	system("paste $test $scrdir/$pref.predict $scrdir/$pref.scores > $scrdir/$pref.tmp");
}
elsif( $bedcheck eq "y" ){
	system("paste $out.bed $test $scrdir/$pref.predict $scrdir/$pref.scores > $scrdir/$pref.tmp");
}
else{
	print "Error: answer to existence of $out.bed was not y or n\n";
	exit;
}
system("$progdir/shorten.pl $scrdir/$pref.tmp $out.predict $thresh");
if( $pval eq "y" ){
	system("$progdir/assign_pvals.pl $out.predict $out.bk.scores $out.predict");
	system("rm $scrdir/$pref.bk.seq");
}
system("rm $scrdir/$pref.log");
system("rm -r $scrdir/$pref\_parts");
system("rm $scrdir/$pref.tmp");
system("rm $scrdir/$pref.scores");

print "\nFinished!\n\n";
exit;

# function that checks $out.log for success of script
# $_[0] is name of script last run
sub check_stat{
	my $line;
	$line = `tail -1 $out.log`;
	chomp($line);
	if( $line ne "success" ){
		print "Error in script $_[0]; exiting program...\n";
		exit;
	}
}

















