#!/usr/bin/perl
# run svm-train with the -v option, testing a varity of C+ and C- parameters for the optimum F-measure

use Getopt::Long;

#
# command line options
#
$input = "";
$output = "";
$progdir = ".";
$kernel = "r";
$stat = "";
$cross = 3;
$options = "Usage: ./search_svm_params.pl <options>
	-t  training data, in SVM format (required)
	-o  output prefix (required)
	-l  log output (status report, optional)
	-p  program directory (optional, default $progdir)
	-k  kernel (l or r, default $kernel)
	-n  n-way cross validation (integer bigger than 2, default $cross)
";

GetOptions('t=s' => \$input,
	   'o=s' => \$output,
	   'p=s' => \$progdir,
	   'k=s' => \$kernel,
	   'l=s' => \$stat,
	   'n=i' => \$cross) || die "$options\n";

if( $input eq "" ){
	print "Error: set -t\n$options\n";
	exit;
}
if( $output eq "" ){
	print "Error: set -o\n$options\n";
	exit;
}
if( $kernel ne "r" && $kernel ne "l" ){
	print "Error: set -k to be l or r\n$options\n";
	exit;
}
if( $cross < 2 ){
	print "Error: set -n to be at least 2\n$options\n";
	exit;
}
if( $stat eq "" ){
	$stat = sprintf("%s.log", $output);
}

# write current status to stat file
open(STAT,">$stat");
print STAT "Starting search\n";

$quiet = 2; 				# set to 1 to suppress printing to stdout, or 2 for extra output
$M = 1; 				# number of times to repeat measuring F-score for each C+,C- combo; 20 is quite robust
$l = int(rand(100)); 			# input to srand, using added -z option
$inc = 2;				# initial power of 2 to increment at
&update_cs( 5, 5, $inc , 11 );
#&update_cs( 1, 1, $inc, 5 );		# uncomment for faster testing
&update_cs( 3, 3, $inc, 9 );		# uncomment for slightly faster testing
&test_svm();

# now refine grid over C+, C-, dividing inc by 2
for($i=0; $i< 2; $i++){
	$inc/=2;
	&update_cs( $bestcpow, $bestnpow, $inc, 5 );
	&test_svm();
}
open(OUT,">$output");
if( $kernel eq "l" ){print OUT "F-value: $avgf[$besti][$bestj] Precision: $avgprec[$besti][$bestj] Specificity: $avgspec[$besti][$bestj] Sensitivity: $avgsens[$besti][$bestj] C+: $bestcp C-: $bestcn\n";}
if( $kernel eq "r" ){print OUT "F-value: $avgf[$besti][$bestj] Precision: $avgprec[$besti][$bestj] Specificity: $avgspec[$besti][$bestj] Sensitivity: $avgsens[$besti][$bestj] C: $bestcp gamma: $bestcn\n";}

system("rm $output.tmp");

print STAT "success\n";
close(STAT);

exit;

#######################################

# function to update the weight parameters; all represent powers of 2, 11 tests each
# USAGE: $_[0]: <middle of C+>, $_[1]: <middle of C->, $_[2]: increment, $_[3]: total (odd number)
sub update_cs{
	my $i;
	my $j;
	my @tmp;
	@cps = ();
	@cns = ();
	@avgf = ();
	@avgsens = ();
	@avgspec = ();
	@avgprec = ();
	$total = $_[3];
	for($i=-($total-1)/2; $i< ($total+1)/2; $i++){
		push( @cps, $_[0]+$_[2]*$i );
		push( @cns, $_[1]+$_[2]*$i );
		@tmp = ();
		for($j=-($total-1)/2; $j< ($total+1)/2; $j++){
			push( @tmp, 0 );
		}
		push( @avgf, [@tmp] );
		push( @avgsens, [@tmp] );
		push( @avgspec, [@tmp] );
		push( @avgprec, [@tmp] );
	}
}

# function to run svm over all options of cps and cpn (no inputs)
# repeat the measurement M times (declared at top), take average
# no return, but write to global values bestcp, bestcn, bestf
sub test_svm{
	my $i;
	my $j;
	my $k;
	if( $quiet != 1 ){
		print "\n   Testing, increment 2^$inc, $M times:\n";
		if( $kernel eq "l" ){
			printf("       C+ from %f to %f (2^%.3f to 2^%.3f)\n", 2**$cps[0], 2**$cps[$#cps], $cps[0], $cps[$#cps]);
			printf("       C- from %f to %f (2^%.3f to 2^%.3f)\n", 2**$cns[0], 2**$cns[$#cns], $cns[0], $cns[$#cns]);
		}
		if( $kernel eq "r" ){
			printf("       C from %f to %f (2^%.3f to 2^%.3f)\n", 2**$cps[0], 2**$cps[$#cps], $cps[0], $cps[$#cps]);
			printf("       gamma from %f to %f (2^%.3f to 2^%.3f)\n", 2**$cns[0], 2**$cns[$#cns], $cns[0], $cns[$#cns]);
		}
	}
	for($k=0; $k< $M; $k++){
		for($i=0; $i<= $#cps; $i++){
			for($j=0; $j<= $#cns; $j++){
				if( $quiet != 1 && $quiet == 2 ){
					if( $kernel eq "l" ){print "\r       iter $k, log_2 C+ = $cps[$i], log_2 C- = $cns[$j]         ";}
					if( $kernel eq "r" ){print "\r       iter $k, log_2 C = $cps[$i], log_2 gamma = $cns[$j]         ";}
				}
				$l++;
				$cp=2**$cps[$i];
				$cn=2**$cns[$j];
				if( $kernel eq "r" ){
					$ker = 2;
				}
				if( $kernel eq "l" ){
					$ker = 0;
				}
				$opts = "-t $ker -v $cross -w1 $cp -w-1 $cn -z $l";
				system("$progdir/svm-train $opts $input > $output.tmp");
				&check_stat("svm-train");
#				$f = `awk '{if(\$1~/F-measure/) print \$2}' < $output.tmp`;
				$sens = `awk '{if(\$1~/Sensitivity/) print \$2}' < $output.tmp`;
				$spec = `awk '{if(\$1~/Specificity/) print \$2}' < $output.tmp`;
				$prec = `awk '{if(\$1~/Precision/) print \$2}' < $output.tmp`;
				chomp($sens);
				chomp($spec);
				chomp($prec);
#				$f = ($spec + $sens)/2; # actually Balanced Accuracy (BAC), but call f 
				if( $prec != 0 || $sens != 0 ){
					$f = 2*$prec*$sens/($prec+$sens); # actual f-measure 
				}
				else{
					$f = 0;
				}
				chomp($f);
				if( $f =~ /nan/ ){
					$f = 0;
					$sens = 0;
					$spec = 0;
					$prec = 0;
				}
				$avgf[$i][$j] += $f/$M;
				$avgsens[$i][$j] += $sens/$M;
				$avgspec[$i][$j] += $spec/$M;
				$avgprec[$i][$j] += $prec/$M;
			}
		}
	}
	$besti=0;
	$bestj=0;
	$bestcp=2**$cps[0];
	$bestcn=2**$cns[0];
	$bestcpow=$cps[0];
	$bestnpow=$cns[0];
	for($i=0; $i< $#cps; $i++){
		for($j=0; $j<= $#cns; $j++){
			if( $avgf[$i][$j] > $avgf[$besti][$bestj] ){
				$besti = $i;
				$bestj = $j;
				$bestcp = 2**$cps[$i];
				$bestcn = 2**$cns[$j];
				$bestcpow = $cps[$i];
				$bestnpow = $cns[$j];
			}
		}
	}
	if( $quiet != 1 ){		
		if( $kernel eq "l" ){printf ("\n   Result: F=%f at C+=%f (2^%.3f), C-=%f (2^%.3f)\n", $avgf[$besti][$bestj], $bestcp, $bestcpow, $bestcn, $bestnpow);}
		if( $kernel eq "r" ){printf ("\n   Result: F=%f at C=%f (2^%.3f), gamma=%f (2^%.3f)\n", $avgf[$besti][$bestj], $bestcp, $bestcpow, $bestcn, $bestnpow);}
	}
}

# function that checks $dir/$pref.log for success of script
# $_[0] is name of script last run
sub check_stat{
        my $line;
        $line = `tail -1 $output.tmp`;
        chomp($line);
        if( $line ne "success" ){
                print "Error in script $_[0]; exiting program...\n";
                exit;
        }
}




